#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <pvm3.h>

#include "cpuinfo.h"
#include "main.h"
#include "misc.h"

#define CPUINFO "/proc/cpuinfo"

static int cpuinfo_send(CpuInfo *info)
{
	assert(info != NULL);

	printf("client: sending cpuinfo\n");

	pvm_initsend(PvmDataDefault);
	pvm_pkint(&(info->num), 1, 1);
	pvm_pkstr(info->vendor_id ? info->vendor_id : "(unknown)");
	pvm_pkstr(info->model_name ? info->model_name : "(unknown)");
	pvm_pkstr(info->cache_size ? info->cache_size : "(unknown)");
	pvm_pkdouble(&(info->mhz), 1, 1);
	pvm_pkdouble(&(info->bogomips), 1, 1);
	pvm_pkstr(info->flags ? info->flags : "");
	pvm_pkint(&(info->cpu_family), 1, 1);
	pvm_pkint(&(info->model), 1, 1);
	pvm_pkint(&(info->stepping), 1, 1);
	pvm_send(pvm_parent(), MSGTAG_CPUINFO);

	return EXIT_SUCCESS;
}

void cpuinfo_free(CpuInfo *info)
{
	assert(info != NULL);

	if(info->vendor_id) free(info->vendor_id);
	if(info->model_name) free(info->model_name);

	free(info);
}

static char *get_val_string(char *line)
{
	char *ptr, *new;
	int len;

	ptr = strstr((const char *)line, ": ");
	if(ptr == NULL)
	{
		return NULL;
	}

	ptr += 2;

	len = strlen(ptr);
	if(ptr[len - 1] == '\n') len --;

	new = malloc((len + 1) * sizeof(char));
	new[len] = '\0';
	assert(new);
	new = strncpy(new, (const char *)ptr, len);

	printf("client: get_val_string: %s\n", new);

	return new;
}

static int get_val_int(char *line)
{
	char *ptr;
	int val;

	ptr = get_val_string(line);
	if(ptr)
	{
		val = (int)strtol((const char *)ptr, NULL, 10);
		free(ptr);
		return val;
	}
	else
	{
		return -1;
	}
}

static double get_val_double(char *line)
{
	char *ptr;
	double val;

	ptr = get_val_string(line);
	if(ptr)
	{
		val = strtod(ptr, NULL);
		free(ptr);
		return val;
	}
	else
	{
		return -1;
	}
}

int cpuinfo(void)
{
	FILE *fp;
	char buf[2048];
	CpuInfo *info = NULL;

	printf("client: running cpuinfo()\n");

	fp = fopen(CPUINFO, "r");
	if(fp)
	{
		printf("client: opened /proc/cpuinfo\n");
		while(fgets(buf, 2048, fp))
		{
			if(strncmp(buf, "processor", 9) == 0)
			{
				if(info)
				{
					/* send previous data */
					cpuinfo_send(info);
					cpuinfo_free(info);
					info = NULL;
				}

				info = m_new0(CpuInfo, 1);
				info->num = get_val_int(buf);

				printf("client: cpu #%d\n", info->num);
			}
			else if(strncmp(buf, "vendor_id", 9) == 0)
			{
				assert(info != NULL);
				info->vendor_id = get_val_string(buf);
				memset(buf, 0, 2048);
			}
			else if((strncmp(buf, "model name", 10) == 0) ||
				(strncmp(buf, "cpu model", 9) == 0)) /* MIPS/IP22 */
			{
				assert(info != NULL);
				info->model_name = get_val_string(buf);
				memset(buf, 0, 2048);
			}
			else if(strncmp(buf, "cpu MHz", 7) == 0)
			{
				assert(info != NULL);
				info->mhz = get_val_double(buf);
			}
			else if(strncmp(buf, "cache size", 10) == 0)
			{
				assert(info != NULL);
				info->cache_size = get_val_string(buf);
			}
			else if((strncmp(buf, "bogomips", 8) == 0) ||
				(strncmp(buf, "BogoMIPS", 8) == 0)) /* MIPS/IP22 */
			{
				assert(info != NULL);
				info->bogomips = get_val_double(buf);
			}
			else if(strncmp(buf, "flags", 5) == 0)
			{
				assert(info != NULL);
				info->flags = get_val_string(buf);
			}
			else if(strncmp(buf, "cpu family", 10) == 0)
			{
				assert(info != NULL);
				info->cpu_family = get_val_int(buf);
			}
			else if(strncmp(buf, "model", 5) == 0)
			{
				assert(info != NULL);
				info->model = get_val_int(buf);
			}
			else if(strncmp(buf, "stepping", 8) == 0)
			{
				assert(info != NULL);
				info->stepping = get_val_int(buf);
			}
		} /* while(fgets()) */

		if(info)
		{
			cpuinfo_send(info);
			cpuinfo_free(info);
			info = NULL;
		}

		fclose(fp);
		return EXIT_SUCCESS;
	}

	return EXIT_FAILURE;
}

