#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <unistd.h>

#include <sys/time.h>
#include <pvm3.h>

#include "main.h"
#include "master.h"
#include "misc.h"
#include "slist.h"
#include "cpuinfo.h"
#include "pciinfo.h"
#include "network.h"
#include "toxml.h"

int master_spawn(const char *cmd, SList **client_list)
{
	int info, i, host_tid, my_tid;
	int nhost, narch; /* number of hosts / architectures */
	struct pvmhostinfo *hostinfo;
	ClientInfo *cinfo;
	SList *clients = NULL;
	char buf[256];

	my_tid = pvm_mytid();

	/* i'm the parent */
	info = pvm_config(&nhost, &narch, &hostinfo);
	if(info < 0)
	{
		printf("%s: master: pvm_config returned %d\n", prog, info);
		pvm_perror((char *)cmd);
		pvm_exit();
		return EXIT_FAILURE;
	}

	host_tid = pvm_tidtohost(my_tid);

	for(i = 0; i < nhost; i ++)
	{
		cinfo = m_new0(ClientInfo, 1);

		cinfo->tid = hostinfo[i].hi_tid;
		cinfo->name = m_strdup(hostinfo[i].hi_name);
		cinfo->master = (host_tid == hostinfo[i].hi_tid);

		printf("%s: master: host %d: %s (%s) TID: %d%s\n", prog, i,
			cinfo->name,
			hostinfo[i].hi_arch,
			cinfo->tid,
			cinfo->master ? " - it's me!" : "");

		memset(buf, 0, 256 * sizeof(char));
		sprintf(buf, "%s/libexec/%s/%s/%s",
			MASTERROOT, BINARYNAME, hostinfo[i].hi_arch, BINARYNAME);

		info = pvm_spawn(buf, NULL, PvmTaskHost, hostinfo[i].hi_name,
			1, &(cinfo->child_tid));

		if(info < 0)
		{
			printf("%s: master: pvm_spawn returned %d\n", prog, info);
			pvm_perror((char *)cmd);
			pvm_exit();
			return EXIT_FAILURE;
		}

		info = pvm_notify(PvmTaskExit, MSGTAG_EXITED, 1, &(cinfo->child_tid));
		if(info < 0)
		{
			printf("%s: master: pvm_notify returned %d\n", prog, info);
			pvm_perror((char *)cmd);
			pvm_exit();
			return EXIT_FAILURE;
		}

		printf("%s: master: spawned task (%d)\n", prog, cinfo->child_tid);

		clients = slist_append(clients, cinfo);
	}

	if(client_list) *client_list = clients;

	return nhost;
}

static ClientInfo *cinfo_from_tid(SList *clients, int tid)
{
	ClientInfo *info = NULL;
	SList *item = clients;

	while(item)
	{
		info = (ClientInfo *)item->data;
		assert(info != NULL);

		if(info->child_tid == tid)
		{
			return info;
		}

		item = item->next;
	}

	return NULL;
}
		
int master_collect(int nhost, SList *clients)
{
	int bufid, child_tid, info, nbytes, msgtag;
	unsigned long int memsize;
	CpuInfo *cinf;
	PciDevInfo *pinfo;
	ClientInfo *client_info;
	BenchmarkResult *bm_result;
	char buf[2048];
	struct timeval tv;

	/* wait for results */
	while(1)
	{
		/* mem info */
		bufid = pvm_nrecv(-1, MSGTAG_MEMINFO);
		if(bufid != 0)
		{
			info = pvm_bufinfo(bufid, &nbytes, &msgtag, &child_tid);

			client_info = cinfo_from_tid(clients, child_tid);

			pvm_upkulong(&memsize, 1, 1);

			if(client_info != NULL)
			{
				client_info->memsize = memsize;
			}
		}

		/* interface info eth0 */
		bufid = pvm_nrecv(-1, MSGTAG_ETH0INFO);
		if(bufid != 0)
		{
			info = pvm_bufinfo(bufid, &nbytes, &msgtag, &child_tid);

			client_info = cinfo_from_tid(clients, child_tid);
			if(client_info != NULL)
			{
				client_info->eth0 = m_new0(NetworkInterfaceInfo, 1);

				/* MAC */
				pvm_upkstr(buf);
				client_info->eth0->hwaddr = m_strdup(buf);

				/* IP */
				pvm_upkbyte(client_info->eth0->ip, 4, 1);
			}
		}

		/* cpu infos */
		bufid = pvm_nrecv(-1, MSGTAG_CPUINFO);
		if(bufid != 0)
		{
			/* got a message */
			info = pvm_bufinfo(bufid, &nbytes, &msgtag, &child_tid);

			client_info = cinfo_from_tid(clients, child_tid);
			if(client_info == NULL)
			{
				printf("%s: master: message from unknown client\n", prog);
				break;
			}

#if 0
			printf("%s: master: got a buffer from TID %d: %d bytes\n",
				prog, child_tid, nbytes);
#endif

			cinf = m_new0(CpuInfo, 1);
			pvm_upkint(&(cinf->num), 1, 1);
			
			pvm_upkstr(buf);
			cinf->vendor_id = m_strdup(buf);
			
			pvm_upkstr(buf);
			cinf->model_name = m_strdup(buf);

			pvm_upkstr(buf);
			cinf->cache_size = m_strdup(buf);
			
			pvm_upkdouble(&(cinf->mhz), 1, 1);
			pvm_upkdouble(&(cinf->bogomips), 1, 1);

			pvm_upkstr(buf);
			cinf->flags = m_strdup(buf);

			pvm_upkint(&(cinf->cpu_family), 1, 1);
			pvm_upkint(&(cinf->model), 1, 1);
			pvm_upkint(&(cinf->stepping), 1, 1);

			printf("%s: #%d: %s %s (%d:%d:%d)\n"
				   "\t\t%.2f MHz (%.2f BogoMIPS)\n"
				   "\t\tcache size: %s\n"
				   "\t\tflags: %s\n"
				   , prog,
				cinf->num, cinf->vendor_id, cinf->model_name,
				cinf->cpu_family, cinf->model, cinf->stepping,
				cinf->mhz, cinf->bogomips, cinf->cache_size,
				cinf->flags);

			if(client_info)
			{
				client_info->cpus = slist_append(client_info->cpus, cinf);
			}
		}

		/* PCI devices */
		bufid = pvm_nrecv(-1, MSGTAG_PCIINFO);
		if(bufid != 0)
		{
			info = pvm_bufinfo(bufid, &nbytes, &msgtag, &child_tid);

			pinfo = m_new0(PciDevInfo, 1);
			
			pvm_upkuint(&(pinfo->c), 1, 1);
			pinfo->c &= 0xFFFF;
			
			pvm_upkuint(&(pinfo->bus), 1, 1);
			pinfo->bus &= 0xFF;
			
			pvm_upkuint(&(pinfo->dev), 1, 1);
			pinfo->dev &= 0xFF;
			
			pvm_upkuint(&(pinfo->func), 1, 1);
			pinfo->func &= 0xFF;
			
			pvm_upkuint(&(pinfo->vendor_id), 1, 1);
			pinfo->vendor_id &= 0xFFFF;
			
			pvm_upkuint(&(pinfo->device_id), 1, 1);
			pinfo->device_id &= 0xFFFF;

			client_info = cinfo_from_tid(clients, child_tid);
			if(client_info != NULL)
			{
				client_info->pcidevs = slist_append(
						client_info->pcidevs, pinfo);
			}
		}

		/* benchmarking results */
		bufid = pvm_nrecv(-1, MSGTAG_BMRESULT);
		if(bufid != 0)
		{
			/* info about buffer */
			info = pvm_bufinfo(bufid, &nbytes, &msgtag, &child_tid);

			/* benchmark name */
			pvm_upkstr(buf);

			/* seconds */
			pvm_upklong(&(tv.tv_sec), 1, 1);

			/* microseconds */
			pvm_upklong(&(tv.tv_usec), 1, 1);

			printf("master: [%s] %ld seconds, %ld microseconds\n",
				buf, tv.tv_sec, tv.tv_usec);

			client_info = cinfo_from_tid(clients, child_tid);
			if(client_info)
			{
				bm_result = m_new0(BenchmarkResult, 1);

				bm_result->name = m_strdup(buf);
				bm_result->sec = tv.tv_sec;
				bm_result->usec = tv.tv_usec;

				client_info->bm_results = slist_append(client_info->bm_results,
					bm_result);
			}
		}

		bufid = pvm_nrecv(-1, MSGTAG_EXITED);
		if(bufid != 0)
		{
			/* one client exited, count down host number */
			pvm_upkint(&child_tid, 1, 1);
			printf("%s: master: child (TID %d) exited\n", prog, child_tid);
			nhost --;

			client_info = cinfo_from_tid(clients, child_tid);

			if(client_info)
			{
				toxml_write_client(client_info);
			}
		}

		if(nhost <= 0) break;

		usleep(100);
	}

	return EXIT_SUCCESS;
}
