#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "main.h"
#include "master.h"
#include "cpuinfo.h"
#include "pciinfo.h"

static FILE *toxml_file = NULL;

int toxml_init(const char *filename, const char *xslfilename)
{
	toxml_file = fopen(filename, "w");
	
	if(toxml_file == NULL)
	{
		fprintf(stderr, "%s, master: could not open '%s' for writing\n",
			prog, filename);

		return EXIT_FAILURE;
	}

	fprintf(toxml_file, "<?xml version=\"1.0\"?>\n");
	if(xslfilename != NULL)
	{
		fprintf(toxml_file,
			"<?xml-stylesheet href=\"%s\" type=\"text/xsl\"?>\n",
			xslfilename);
	}
	fprintf(toxml_file, "<cluster-info>\n");

	return EXIT_SUCCESS;
}

void toxml_close(void)
{
	if(toxml_file != NULL)
	{
		fprintf(toxml_file, "</cluster-info>\n");
		fclose(toxml_file);
	}
}

static void write_str(const char *tag, const char *indent, char *value)
{
	fprintf(toxml_file, "%s<%s>%s</%s>\n", indent, tag, value, tag);
}

static void write_int(const char *tag, const char *indent, int value)
{
	fprintf(toxml_file, "%s<%s>%d</%s>\n", indent, tag, value, tag);
}

static void write_uix(const char *tag, const char *indent, unsigned int value)
{
	fprintf(toxml_file, "%s<%s>0x%4.4X</%s>\n", indent, tag, value, tag);
}

static void write_ulg(const char *tag, const char *indent, unsigned long value)
{
		    fprintf(toxml_file, "%s<%s>%lu</%s>\n", indent, tag, value, tag);
}

static void write_dbl(const char *tag, const char *indent, double value)
{
		    fprintf(toxml_file, "%s<%s>%.2f</%s>\n", indent, tag, value, tag);
}

int toxml_write_client(ClientInfo *client_info)
{
	SList *item;
	CpuInfo *cpuinfo;
	BenchmarkResult *bm_result;
	PciDevInfo *pinfo;
	char *startp, *bufp, flag[256];

	if(toxml_file == NULL) return EXIT_FAILURE;

	fprintf(toxml_file, "  <client>\n");
	write_str("hostname", "    ", client_info->name);
	write_ulg("memsize", "    ", client_info->memsize);
	
	fprintf(toxml_file, "    <cpus>\n");
	item = client_info->cpus;
	while(item)
	{
		cpuinfo = (CpuInfo *)item->data;
		assert(cpuinfo != NULL);

		fprintf(toxml_file, "      <cpu>\n");
		write_int("num", "        ", cpuinfo->num);
		write_str("vendor_id", "        ", cpuinfo->vendor_id);
		write_str("model_name", "        ", cpuinfo->model_name);
		write_str("cache_size", "        ", cpuinfo->cache_size);
		write_dbl("mhz", "        ", cpuinfo->mhz);
		write_dbl("bogomips", "        ", cpuinfo->bogomips);
		fprintf(toxml_file, "        <flags>\n");
		startp = bufp = cpuinfo->flags;
		while(bufp && bufp != '\0')
		{
			while((*bufp != '\0') && (*bufp != ' ')) bufp ++;

			memset(flag, 0, 256 * sizeof(char));
			strncpy(flag, startp, (bufp - startp));

			fprintf(toxml_file, "          <flag_%s/>\n", flag);

			if(*bufp == '\0') { bufp = NULL; }
			else { bufp++; startp = bufp; }
		}
		
		fprintf(toxml_file, "        </flags>\n");
		fprintf(toxml_file, "        <version>\n");
		write_int("family", "          ", cpuinfo->cpu_family);
		write_int("model", "          ", cpuinfo->model);
		write_int("stepping", "          ", cpuinfo->stepping);
		fprintf(toxml_file, "        </version>\n");
		fprintf(toxml_file, "      </cpu>\n");
		
		item = item->next;
	}
	fprintf(toxml_file, "    </cpus>\n");

	/* network information */
	if(client_info->eth0)
	{
		fprintf(toxml_file, "    <network>\n");
		write_str("hwaddr", "      ", client_info->eth0->hwaddr);
		fprintf(toxml_file, "    </network>\n");
	}

	/* PCI devices */
	fprintf(toxml_file, "    <devices>\n");
	item = client_info->pcidevs;
	while(item)
	{
		pinfo = (PciDevInfo *)item->data;

		fprintf(toxml_file, "      <pci_device>\n");
		write_uix("class", "        ", pinfo->c);
		write_uix("bus", "        ", pinfo->bus);
		write_uix("dev", "        ", pinfo->dev);
		write_uix("function", "        ", pinfo->func);
		write_uix("vendor_id", "        ", pinfo->vendor_id);
		write_uix("device_id", "        ", pinfo->device_id);
		fprintf(toxml_file, "      </pci_device>\n");

		item = item->next;
	}
	fprintf(toxml_file, "    </devices>\n");

	/* benchmarking results */
	if(client_info->bm_results)
	{
		fprintf(toxml_file, "    <benchmarks>\n");
		item = client_info->bm_results;
		while(item)
		{
			bm_result = (BenchmarkResult *)item->data;

			fprintf(toxml_file, "      <result>\n");
			write_str("test", "        ", bm_result->name);
			write_int("seconds", "        ", bm_result->sec);
			write_int("useconds", "        ", bm_result->usec);
			fprintf(toxml_file, "      </result>\n");
			item = item->next;
		}
		fprintf(toxml_file, "    </benchmarks>\n");
	}
	
	fprintf(toxml_file, "  </client>\n");
	fflush(toxml_file);

	return EXIT_SUCCESS;
}
